package com.tillekesoft.accessibilityservices;

import android.accessibilityservice.AccessibilityService;
import android.accessibilityservice.AccessibilityServiceInfo;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.pm.ActivityInfo;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.view.accessibility.AccessibilityEvent;
import android.view.accessibility.AccessibilityManager;
import android.view.accessibility.AccessibilityNodeInfo;

import java.util.ArrayList;
import java.util.List;
import android.accessibilityservice.GestureDescription;
import android.graphics.Path;
import android.os.Handler;

import anywheresoftware.b4a.BA;

public class AccessibilityEventsListenerWrapper extends AccessibilityService {
    public AccessibilityEventsListener acs;
    public static boolean isServiceActive = false;
    private ArrayList<AccessibilityNodeInfo> MyViewNodes;
    private ArrayList<Object> infoNodes;

    @Override
    public void onCreate() {
        startService(new Intent(this, this.getClass())); //Important if the service will be bound instead of started
    }

    @Override
    public void onServiceConnected() {
        isServiceActive = true;
    }

    @Override
    public void onDestroy() {
        isServiceActive = false;
    }

    private String getEventType(AccessibilityEvent event) {
        switch (event.getEventType()) {
            case AccessibilityEvent.TYPE_NOTIFICATION_STATE_CHANGED:
                return "TYPE_NOTIFICATION_STATE_CHANGED";
            case AccessibilityEvent.TYPE_VIEW_CLICKED:
                return "TYPE_VIEW_CLICKED";
            case AccessibilityEvent.TYPE_VIEW_FOCUSED:
                return "TYPE_VIEW_FOCUSED";
            case AccessibilityEvent.TYPE_VIEW_LONG_CLICKED:
                return "TYPE_VIEW_LONG_CLICKED";
            case AccessibilityEvent.TYPE_VIEW_SELECTED:
                return "TYPE_VIEW_SELECTED";
            case AccessibilityEvent.TYPE_WINDOW_STATE_CHANGED:
                return "TYPE_WINDOW_STATE_CHANGED";
            case AccessibilityEvent.TYPE_VIEW_TEXT_CHANGED:
                return "TYPE_VIEW_TEXT_CHANGED";
            case AccessibilityEvent.TYPE_VIEW_HOVER_ENTER:
                return "TYPE_VIEW_HOVER_ENTER";
            case AccessibilityEvent.TYPE_VIEW_HOVER_EXIT:
                return "TYPE_VIEW_HOVER_EXIT";
            case AccessibilityEvent.TYPE_TOUCH_EXPLORATION_GESTURE_START:
                return "TYPE_TOUCH_EXPLORATION_GESTURE_START";
            case AccessibilityEvent.TYPE_TOUCH_EXPLORATION_GESTURE_END:
                return "TYPE_TOUCH_EXPLORATION_GESTURE_END";
            case AccessibilityEvent.TYPE_WINDOW_CONTENT_CHANGED:
                return "TYPE_WINDOW_CONTENT_CHANGED";
            case AccessibilityEvent.TYPE_VIEW_SCROLLED:
                return "TYPE_VIEW_SCROLLED";
            case AccessibilityEvent.TYPE_VIEW_TEXT_SELECTION_CHANGED:
                return "TYPE_VIEW_TEXT_SELECTION_CHANGED";
            case AccessibilityEvent.TYPE_ANNOUNCEMENT:
                return "TYPE_ANNOUNCEMENT";
            case AccessibilityEvent.TYPE_VIEW_ACCESSIBILITY_FOCUSED:
                return "TYPE_VIEW_ACCESSIBILITY_FOCUSED";
            case AccessibilityEvent.TYPE_VIEW_ACCESSIBILITY_FOCUS_CLEARED:
                return "TYPE_VIEW_ACCESSIBILITY_FOCUS_CLEARED";
            case AccessibilityEvent.TYPE_VIEW_TEXT_TRAVERSED_AT_MOVEMENT_GRANULARITY:
                return "TYPE_VIEW_TEXT_TRAVERSED_AT_MOVEMENT_GRANULARITY";
            case AccessibilityEvent.TYPE_GESTURE_DETECTION_START:
                return "TYPE_GESTURE_DETECTION_START";
            case AccessibilityEvent.TYPE_GESTURE_DETECTION_END:
                return "TYPE_GESTURE_DETECTION_END";
            case AccessibilityEvent.TYPE_TOUCH_INTERACTION_START:
                return "TYPE_TOUCH_INTERACTION_START";
            case AccessibilityEvent.TYPE_TOUCH_INTERACTION_END:
                return "TYPE_TOUCH_INTERACTION_END";
            case AccessibilityEvent.TYPE_WINDOWS_CHANGED:
                return "TYPE_WINDOWS_CHANGED";
            case AccessibilityEvent.TYPE_VIEW_CONTEXT_CLICKED:
                return "TYPE_VIEW_CONTEXT_CLICKED";
            case AccessibilityEvent.TYPE_ASSIST_READING_CONTEXT:
                return "TYPE_ASSIST_READING_CONTEXT";
        }
        return "default";
    }

    private void findChildViews(AccessibilityNodeInfo parentView) {
        if (parentView == null || parentView.getClassName() == null) {
            return;
        }
        int childCount = parentView.getChildCount();
        if (childCount == 0) {
            MyViewNodes.add(parentView);
        } else {
            for (int i = 0; i < childCount; i++) {
                findChildViews(parentView.getChild(i));
            }
        }
    }

    public static void logViewHierarchy(AccessibilityNodeInfo nodeInfo, final int depth) {
        if (nodeInfo == null) return;
        String spacerString = "";
        for (int i = 0; i < depth; ++i) {
            spacerString += '-';
        }
        for (int i = 0; i < nodeInfo.getChildCount(); ++i) {
            logViewHierarchy(nodeInfo.getChild(i), depth + 1);
        }
    }

    public static boolean isAccessibilityEnabled(Context context, String id) {
        AccessibilityManager am = (AccessibilityManager) context
                .getSystemService(Context.ACCESSIBILITY_SERVICE);
        List<AccessibilityServiceInfo> runningServices = null;
        if (am != null) {
            runningServices = am
                    .getEnabledAccessibilityServiceList(AccessibilityEvent.TYPES_ALL_MASK);
            for (AccessibilityServiceInfo service : runningServices) {
                if (id.equals(service.getId())) {
                    return true;
                }
            }
        }
        return false;
    }

    private ActivityInfo getActivityInfo(ComponentName componentName) {
        try {
            return getPackageManager().getActivityInfo(componentName, 0);
        } catch (PackageManager.NameNotFoundException e) {
            return null;
        }
    }

    @Override
    public void onAccessibilityEvent(AccessibilityEvent event) {
        if(acs != null){
            MyViewNodes = new ArrayList<>();
            infoNodes = new ArrayList<>();
//            AccessibilityNodeInfo node = event.getSource();
            AccessibilityNodeInfo node = getRootInActiveWindow();
            findChildViews(node);
            infoNodes.add(MyViewNodes);
            acs.ba.raiseEvent(this,acs.eventName + "_onaccessibilityevent",new Object[]{event, infoNodes});

            ComponentName componentName = new ComponentName(event.getPackageName().toString(), event.getClassName().toString());
            ActivityInfo activityInfo = getActivityInfo(componentName);
            boolean isActivity = activityInfo != null;

            if (isActivity) {
                String activityName = componentName.flattenToShortString();
                acs.ba.raiseEvent(this, acs.eventName + "_onactivitynameretrieved", new Object[]{activityName});
                return;
            }
        }
    }

    @Override
    public void onInterrupt() {
    }


    /**
     * Accessibility...
     */
    @BA.Version(0.10F)
    @BA.Author("Moster67 aka Mikael Osterhed")
    @BA.Events(values = {"OnAccessibilityEvent (Event as Object, Node as Object)", "OnSearchDataReturned (Data as String)", "OnActivityNameRetrieved (Data as String)"})
    @BA.Permissions(values = {"android.permission.BIND_ACCESSIBILITY_SERVICE"})
    @BA.ShortName("AccessibilityEventsListener")
    public static class AccessibilityEventsListener {
        BA ba;
        String eventName;
        private AccessibilityEventsListenerWrapper owner;
		
        public void initialize(final BA ba, String EventName) {
            this.ba = ba;
            this.eventName = EventName.toLowerCase(BA.cul);
            owner = (AccessibilityEventsListenerWrapper) ba.service; //reference to the service
            owner.acs = this;
        }

        public boolean IsServiceActive() {
            if (!isServiceActive) {
                return false;
            }
            return true;
        }

        public void Search(Boolean UseViewById, String criteria) {
            AccessibilityNodeInfo rootNode = owner.getRootInActiveWindow();
            List<AccessibilityNodeInfo> list;

            if (UseViewById) {
                list = rootNode.findAccessibilityNodeInfosByViewId(criteria); //SearchByViewId
            } else {
                list = rootNode.findAccessibilityNodeInfosByText(criteria); //SearchByText
            }

            if (list == null || list.isEmpty()) {
                BA.Log("Search criteria was not found");
                return;
            }

            AccessibilityNodeInfo node = list.get(0);
            ba.raiseEvent(this, eventName + "_onsearchdatareturned", new Object[]{list.get(0).toString()});
        }

        private List SearchResult(Boolean UseViewById, String criteria) {
            AccessibilityNodeInfo rootNode = owner.getRootInActiveWindow();
            List<AccessibilityNodeInfo> list;

            if (UseViewById) {
                list = rootNode.findAccessibilityNodeInfosByViewId(criteria); //SearchByViewId
            } else {
                list = rootNode.findAccessibilityNodeInfosByText(criteria); //SearchByText
            }

            if (list == null || list.isEmpty()) {
                BA.Log("Search criteria was not found");
                return null;
            }
            return list;
        }

        public boolean PerformGlobalAction(int GlobalAction) {
            return owner.performGlobalAction(GlobalAction);
        }

		//Try the dispatchgesture method. The method returned true, but the expected gesture did not occur! Why??
        public boolean dispatchGestureClick() {
			GestureDescription.Builder gestureBuilder = new GestureDescription.Builder();
			Path path= new Path();
			path.moveTo(300, 300);
			gestureBuilder.addStroke(new GestureDescription.StrokeDescription(path, 0, 50));
			return owner.dispatchGesture(gestureBuilder.build(), null, null);
        }

        /**
        *generic method for performing node actions (only partially implemented)
        */
        public boolean PerformNodeActionOnViewWithArgs(Boolean UseViewById, String Criteria, int NodeAction, anywheresoftware.b4a.objects.collections.Map ArgsMap) {
            AccessibilityNodeInfo node = owner.getRootInActiveWindow();
            List<AccessibilityNodeInfo> list1;
            if (!UseViewById) {
                list1 = node.findAccessibilityNodeInfosByText(Criteria); //SearchByText
            } else {
                list1 = node.findAccessibilityNodeInfosByViewId(Criteria); //SearchById
            }
            if (list1 == null || list1.isEmpty()) {
                return false;
            } else {
                AccessibilityNodeInfo node1 = list1.get(0);
                switch (NodeAction) {
                    case AccessibilityConstants.ACTION_SET_TEXT: //2097152: //
                        java.util.Map textMap = B4AMapToJMap(ArgsMap);
                        String theTextValue = "";
                        for (Object value : textMap.values()){
                            //iterate over values
                            theTextValue = value.toString();
                        }
                        Bundle arguments = new Bundle();
                        arguments.putCharSequence(AccessibilityNodeInfo.ACTION_ARGUMENT_SET_TEXT_CHARSEQUENCE, theTextValue);
                        node1.performAction(AccessibilityNodeInfo.ACTION_SET_TEXT, arguments);
                        break;
                    default:
                        node1.performAction(NodeAction);
                        break;
                }
                return true;
            }
        }

         //some public methods (not all are tested). Here is a good place to add other public methods.
         //note that some public methods can be achieved using the PerformNodeActionOnViewWithArgs-method
        public boolean SetText(Boolean UseViewById, String Criteria, CharSequence text){
            List<AccessibilityNodeInfo> list = SearchResult(UseViewById,Criteria);
            if (list == null || list.isEmpty()) {
                BA.Log("In SetText, search criteria was not found");
                return false;
            } else {
                AccessibilityNodeInfo node = list.get(0);
                Bundle arguments = new Bundle();
                arguments.putCharSequence(AccessibilityNodeInfo.ACTION_ARGUMENT_SET_TEXT_CHARSEQUENCE, text);
                node.performAction(AccessibilityNodeInfo.ACTION_SET_TEXT, arguments);
                return true;
            }
        }

        public String GetText(Boolean UseViewById, String Criteria){
            List<AccessibilityNodeInfo> list = SearchResult(UseViewById,Criteria);
            if (list == null || list.isEmpty()) {
                BA.Log("In SetText, search criteria was not found");
                return "";
            } else {
                AccessibilityNodeInfo node = list.get(0);
                return (String) node.getText();
            }
        }

        public boolean SetTextSelection(Boolean UseViewById, String Criteria, int Start, int End){
            List<AccessibilityNodeInfo> list = SearchResult(UseViewById,Criteria);
            if (list == null || list.isEmpty()) {
                BA.Log("In SetText, search criteria was not found");
                return false;
            } else {
                AccessibilityNodeInfo node = list.get(0);
                node.setTextSelection(Start, End);
                return true;
            }
        }

        public int GetTextSelectionStart(Boolean UseViewById, String Criteria){
            List<AccessibilityNodeInfo> list = SearchResult(UseViewById,Criteria);
            if (list == null || list.isEmpty()) {
                BA.Log("In SetText, search criteria was not found");
                return 0;
            } else {
                AccessibilityNodeInfo node = list.get(0);
                return node.getTextSelectionStart();
            }
        }

        public int GetTextSelectionEnd(Boolean UseViewById, String Criteria){
            List<AccessibilityNodeInfo> list = SearchResult(UseViewById,Criteria);
            if (list == null || list.isEmpty()) {
                BA.Log("In SetText, search criteria was not found");
                return 0;
            } else {
                AccessibilityNodeInfo node = list.get(0);
                return node.getTextSelectionEnd();
            }
        }


        //other methods...
        private static void logInstalledAccessiblityServices(Context context) {
            AccessibilityManager am = (AccessibilityManager) context
                    .getSystemService(Context.ACCESSIBILITY_SERVICE);
            List<AccessibilityServiceInfo> runningServices = null;
            if (am != null) {
                runningServices = am
                        .getInstalledAccessibilityServiceList();
                for (AccessibilityServiceInfo service : runningServices) {
                    BA.Log("Service id is " + service.getId());
                }
            } else {
                BA.Log("No service id:s were found");
            }
        }

        public boolean IsAccessibilityServiceEnabled(String id) {
            logInstalledAccessiblityServices(owner); //left to log all accessibility services which are available
            return isAccessibilityEnabled(owner, id);
        }

        public static void delay(int ms) {
            try {
                Thread.sleep(ms);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

        @SuppressWarnings("unchecked")
        java.util.Map<String, Object> B4AMapToJMap(anywheresoftware.b4a.objects.collections.Map map) {
            return (java.util.Map) map.getObject();
        }

        anywheresoftware.b4a.objects.collections.Map JMapToB4AMap(java.util.Map<String, Object> map) {
            anywheresoftware.b4a.objects.collections.Map m = new anywheresoftware.b4a.objects.collections.Map();
            m.Initialize();
            m.getObject().putAll(map);
            return m;
        }
    }
}
